"""
Transect marker tool widget.

Manages having multiple markers on screen at once and the uploading of previous projects.
"""

import kivy.uix as ui
from kivy.uix.button import Button
from kivy.uix.popup import Popup
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.core.window import Window
from os.path import exists
import json
import nccut.functions as func
from nccut.marker import Marker
from nccut.markerwidth import MarkerWidth
from nccut.plotpopup import PlotPopup


class Click:
    """
    Object that mimics a user click.

    Attributes:
        x (float): X coordinate of click point
        y (float): Y coordinate of click point
        pos (tuple): 2 element tuple: (X coord, Y coord)
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.pos = (x, y)


def correct_test(data):
    """
    Check if dictionary has necessary fields to be a marker

    Args:
        data (dict): Dictionary to be tested.

    Returns:
        Boolean, whether dictionary has necessary keys with a list has the value
    """
    keys = list(data.keys())
    if len(keys) == 0:
        return False
    else:
        need = ['Click X', 'Click Y', 'Width']
        for item in need:
            if item not in keys or not isinstance(data[item], list):
                return False
    return True


def marker_find(data, res):
    """
    Recursively examines dictionary and finds marker click coordinates and transect widths.

    Args:
        data (dict): Dictionary to examine
        res (list): Empty list to fill with marker click coordinates and transect widths

    Returns:
        Nested List. A list containing a list for each marker which each contains three lists:
        click X coords, click y coords, and the transect width for each click point in the marker.
        If no qualifying data was found returns empty list. If duplicate data is found (ex: multiple
        variables in a file) only returns one instance of marker data.
    """
    for key in list(data.keys()):
        if key[0:6] == 'Marker':
            if correct_test(data[key]):  # Marker dict has necessary fields
                if len(res) == 0:  # If res empty, always add marker data
                    res.append([data[key]['Click X'], data[key]['Click Y'], data[key]['Width']])
                else:  # If res not empty, ensure found marker data isn't already in res
                    new = True
                    for item in res:
                        l1 = data[key]['Click X']
                        l2 = item[0]
                        if len(l1) == len(l2) and len(l1) == sum([1 for i, j in zip(l1, l2) if i == j]):
                            new = False
                    if new:
                        res.append([data[key]['Click X'], data[key]['Click Y'], data[key]['Width']])
        else:
            if type(data[key]) is dict:  # Can still go further in nested dictionary tree
                marker_find(data[key], res)
            else:
                return res
    return res


class MultiMarker(ui.widget.Widget):
    """
    Transect marker tool widget.

    Created when Transect Marker button is selected. From there on this object manages the creation,
    modification, and data packaging of markers. Manages the uploading of previous projects
    into the viewer.

    Attributes:
        m_on (bool): Whether there are any markers active
        upload_fail (bool): If anything has gone wrong in the project uploading process
        home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        dbtn: RoundedButton, Plot button to activate PlotPopup
        dragging (bool): Whether viewer is in dragging mode
        width_w: :class:`nccut.markerwidth.MarkerWidth` widget to allow for adjustable marker widths
        clicks (int): Number of clicks made by user. Does not decrease when points are deleted
            unless all points are deleted in which case it goes back to zero.
        up_btn: RoundedButton, Upload button for uploading a past project
        nbtn: RoundedButton, New marker button
        plotting: :class:`nccut.plotpopup.PlotPopup`, reference to plotting menu when opened
        curr_width (int): Current marker width being used. Used to initialize width of new markers.
    """
    def __init__(self, home, **kwargs):
        """
        Defines sidebar elements and initializes widget

        Args:
            home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        """
        super(MultiMarker, self).__init__(**kwargs)
        self.m_on = False
        self.upload_fail = False
        self.home = home
        self.dbtn = func.RoundedButton(text="Plot", size_hint=(1, 0.1), font_size=self.home.font)
        self.dbtn.bind(on_press=lambda x: self.gather_popup())
        self.dragging = False
        self.width_w = MarkerWidth(self, size_hint=(1, 0.1))
        self.clicks = 0
        self.plotting = None
        self.curr_width = 40

        # Upload Button
        self.upbtn = func.RoundedButton(text="Upload Project", size_hint=(1, 0.1), font_size=self.home.font)
        self.upbtn.bind(on_press=lambda x: self.upload_pop())
        self.home.ids.sidebar.add_widget(self.upbtn, 1)

        # New Line Button
        self.nbtn = func.RoundedButton(text="New Line", size_hint=(1, 0.1), font_size=self.home.font)

        self.nbtn.bind(on_press=lambda x: self.new_line())
        self.home.ids.sidebar.add_widget(self.nbtn, 1)

    def font_adapt(self, font):
        """
        Updates font of sidebar elements.
        Args:
            font (float): New font size
        """
        self.dbtn.font_size = font
        self.upbtn.font_size = font
        self.nbtn.font_size = font
        self.width_w.font_adapt(font)

    def update_l_col(self, color):
        """
        Asks each marker to update their line color

        Args:
            color (str): New color value: 'Blue', 'Green' or 'Orange'
        """
        for m in self.children:
            m.update_l_col(color)

    def update_c_size(self, value):
        """
       Asks each marker to update their circle size

       Args:
           value (float): New circle size
       """
        for m in self.children:
            m.update_c_size(value)

    def change_dragging(self, val):
        """
        Change whether in dragging mode or not.

        Args:
            val (bool): Whether in dragging mode or not.
        """
        self.dragging = val

    def upload_pop(self):
        """
        Opens popup to ask for name of project file user wishes to upload.
        """
        content = ui.boxlayout.BoxLayout(orientation='horizontal')
        popup = Popup(title="File Name", content=content, size_hint=(0.5, 0.15))
        txt = TextInput(size_hint=(0.7, 1), hint_text="Enter File Name")
        content.add_widget(txt)
        go = Button(text="Ok", size_hint=(0.1, 1))
        go.bind(on_release=lambda x: self.check_file(txt.text, popup))
        close = Button(text="Close", size_hint=(0.2, 1))
        close.bind(on_press=popup.dismiss)
        content.add_widget(go)
        content.add_widget(close)
        popup.open()

    def check_file(self, file, popup):
        """
        Checks is given file name is a properly formatted project file. If it is it uploads
        file. If not, closes popups and shows error message.

        Args:
            file (str): File path
            popup: kivy.uix.popup.Popup, File name input popup (so can close if file invalid)
        """
        if exists(file):
            if file[-5:] == ".json":
                data = json.load(open(file))
                found = marker_find(data, [])
                if len(found) >= 1:
                    popup.dismiss()
                    self.upload_data(found)
                else:
                    content = Label(text="Incorrect File Format")
                    popup2 = Popup(title="Error", content=content, size_hint=(0.5, 0.15))
                    popup2.open()
            else:
                content = Label(text="Incorrect File Format")
                popup2 = Popup(title="Error", content=content, size_hint=(0.5, 0.15))
                popup2.open()
        else:
            content = Label(text="File Not Found")
            popup2 = Popup(title="Error", content=content, size_hint=(0.5, 0.15))
            popup2.open()

    def upload_fail_alert(self):
        """
        Indicate upload has failed
        """
        self.upload_fail = True

    def upload_data(self, points):
        """
        Loads markers from project file.

        Adds markers by 'clicking' the points in the file with the marker width denoted by the file

        Args:
            points: Properly formatted nested list from :class:`nccut.multimarker.marker_find()` function.
        """
        self.upload_fail = False
        if len(self.children) != 0:  # If markers already exist in viewer
            self.children[0].stop_drawing()
            if self.children[0].clicks < 2:  # If any existing markers are incomplete, remove them
                if self.children[0].clicks == 1:
                    self.children[0].del_point()
                self.remove_widget(self.children[0])
        for m in range(0, len(points)):
            marker = Marker(home=self.home, width=self.curr_width)
            clicks = tuple(zip(points[m][0], points[m][1], points[m][2]))
            marker.upload_mode(True)
            self.add_widget(marker)
            for i in clicks:
                touch = Click(i[0], i[1])
                marker.t_width = i[2]
                marker.on_touch_down(touch)
                self.clicks += 1
            marker.upload_mode(False)
            if self.upload_fail:  # If upload goes wrong, stop and undo everything
                self.undo_upload(m)
                return
        self.new_line()

    def undo_upload(self, markers):
        """
        Remove any previous markers that had been uploaded if upload fails

        Args:
            markers (int): Number of markers added so far
        """
        for m in range(0, markers + 1):
            Window.unbind(mouse_pos=self.children[0].draw_line)
            self.remove_widget(self.children[0])
        if len(self.children) == 0:
            # Remove sidebar buttons if deleted marker was the only marker
            self.clicks = 0
            if self.dbtn in self.home.display.current:
                self.home.display.current.remove(self.dbtn)
            if self.width_w in self.home.display.current:
                self.home.display.current.remove(self.width_w)
            if self.dragging:
                self.home.display.drag_mode()
            self.new_line()
        content = Label(text="Project File Markers out of Bounds")
        popup = Popup(title="Error", content=content, size_hint=(0.5, 0.15))
        popup.open()

    def update_width(self, num):
        """
        Update width of active marker.

        Args:
            num (int): New width value
        """
        self.curr_width = num
        self.children[0].update_width(num)

    def del_line(self):
        """
        Delete most recent marker with some safeguards.

        If only one marker on screen, delete but add new marker and remove sidebar elements.
        Otherwise delete current marker and go to previous. If no markers are on screen nothing
        happens.
        """
        if len(self.children) == 0:
            # If no markers on screen do nothing
            return
        Window.unbind(mouse_pos=self.children[0].draw_line)
        self.remove_widget(self.children[0])
        if len(self.children) == 0:
            # Remove sidebar buttons if deleted marker was the only marker
            self.clicks = 0
            if self.dbtn in self.home.display.current:
                self.home.display.current.remove(self.dbtn)
            if self.width_w in self.home.display.current:
                self.home.display.current.remove(self.width_w)
            self.new_line()

    def del_point(self):
        """
        Delete most recently clicked point with some safeguards.

        If no markers are on screen does nothing. If only one marker exists and no points have been clicked,
        does nothing. If more than one marker exists and no points have been clicked on most recent marker,
        Deletes most recent marker and then deletes last point of previous marker. Any other conditions
        simply deletes last clicked point.
        """
        if len(self.children) == 0:
            # If no markers on screen do nothing
            return
        elif self.children[0].clicks == 0:
            if len(self.children) > 1:
                # If no clicks on current marker and not the only marker delete current marker
                self.remove_widget(self.children[0])
            else:
                return
        # Delete point from current marker
        self.children[0].del_point()

    def new_line(self):
        """
        Creates a new marker if not in dragging or editing mode and current marker has at least two clicks.
        """
        if not self.dragging or self.home.display.editing:
            if len(self.children) == 0 or self.children[0].clicks >= 2:
                if len(self.children) != 0:
                    self.children[0].stop_drawing()
                m = Marker(home=self.home, width=self.curr_width)
                self.add_widget(m)

    def gather_popup(self):
        """
        Gather data from markers and call for :class:`nccut.plotpopup.PlotPopup`
        """
        frames = {}
        c = 1
        for i in reversed(self.children):
            if i.clicks > 0:  # Ignore empty markers
                data = {}
                data['Click X'], data['Click Y'], data['Width'] = map(list, zip(*i.points))
                count = 1
                for j in i.base.lines:
                    data["Cut " + str(count)] = j.line.points
                    count += 1
                frames["Marker " + str(c)] = data
                c += 1
        self.plotting = PlotPopup(frames, self.home, self.home.display.config)

    def on_touch_down(self, touch):
        """
        Manages when sidebar elements are added to sidebar and clears them as needed.

        Args:
            touch: MouseMotionEvent, see kivy docs for details
        """
        if not self.dragging:
            if self.home.ids.view.collide_point(*self.home.ids.view.to_widget(*self.to_window(*touch.pos))):
                self.clicks += 1
                if self.clicks >= 1 and self.width_w.parent is None:
                    self.home.ids.sidebar.add_widget(self.width_w, 1)
                if self.clicks >= 2 and self.dbtn.parent is None:
                    self.home.ids.sidebar.add_widget(self.dbtn, 1)
                # If no current marker, create marker. Otherwise, pass touch to current marker.
                if not self.m_on:
                    self.new_line()
                    self.m_on = True
                self.children[0].on_touch_down(touch)



def project_file_printer(data, tab):
    """
    Recursively prints the contents of an output data file from NcCut in a readable fashion.

    Attributes:
        data (dict): Dictionary output from NcCut
        tab (int): Always use 0, used in the recursive process for indentation
    """
    for key in list(data.keys()):
        print(" " * tab, end="")
        if isinstance(data[key], list):
            print(key + ": [" + str(data[key][0]) + ", " + str(data[key][1]) + ", ..." + "]")
        else:
            print(key + ":")
            project_file_printer(data[key], tab + 4)



"""
Singular marker widget.

Graphics and functionality of a singular marker created by the transect marker tool.
"""

import kivy.uix as ui
from kivy.metrics import dp
from kivy.graphics import Color, Ellipse, Line
from kivy.uix.label import Label
import math
import numpy as np
from kivy.core.window import Window
import nccut.functions as functions
from nccut.singletransect import SingleTransect
from nccut.multitransect import MultiTransect


class Marker(ui.widget.Widget):
    """
    Singular marker widget.

    Graphics and functionality of a singular marker created by the transect marker tool.
    Determines endpoints of where transects should be made orthogonally to the user marked
    out line, and then uses a MultiTransect object to manage the transects.

    Attributes:
        clicks (int): Number of clicks user has made. Decreases when points are deleted.
        points (list): List of Tuples, For each click user makes: (X-coord, Y-coord, t_width).
        t_width (int): Current width in pixels of orthogonal transects
        home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        base: :class:`nccut.multitransect.MultiTransect` object that manages transects
        curr_line: kivy.graphics.Line, Line between cursor and last clicked
        number: kivy.uix.label.Label, Reference to the number label
        size: 2 element array of ints, Size of widget
        pos: 2 element array of ints, Position of widget
        l_color: kivy.graphics.Color, Color to use for graphics
        c_size: 2 element tuple of floats that defines size of circles
        line_width (float): Width of lines
    """
    def __init__(self, home, width, **kwargs):
        """
        Sets initial settings and initializes object.

        Args:
            home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
            width (int): Initial transect width to use.
        """
        super(Marker, self).__init__(**kwargs)
        self.clicks = 0
        self.points = []
        self.t_width = width
        self.uploaded = False
        self.home = home
        self.base = MultiTransect(home=self.home)
        self.curr_line = Line()
        self.number = None
        self.size = self.home.display.size
        self.pos = self.home.display.pos
        color = self.home.display.l_col
        if color == "Blue":
            self.l_color = Color(0.28, 0.62, 0.86)
        elif color == "Green":
            self.l_color = Color(0.39, 0.78, 0.47)
        elif color == "Orange":
            self.l_color = Color(0.74, 0.42, 0.13)
        size = self.home.display.cir_size
        self.c_size = (dp(size), dp(size))
        self.line_width = dp(size / 5)

    def update_l_col(self, color):
        """
        Update the line color and redraw all items on canvas.

        Args:
            color (str): New line color to use
        """
        if color == "Blue":
            self.l_color = Color(0.28, 0.62, 0.86)
        elif color == "Green":
            self.l_color = Color(0.39, 0.78, 0.47)
        elif color == "Orange":
            self.l_color = Color(0.74, 0.42, 0.13)
        groups = []
        for c in range(1, self.clicks + 1):
            group = self.canvas.get_group(str(c))
            groups.append(group)
        self.canvas.clear()
        self.remove_widget(self.number)
        self.canvas.add(self.l_color)
        for g in groups:
            for i in g:
                self.canvas.add(i)
        self.canvas.add(self.curr_line)
        if self.clicks > 0:
            self.add_widget(self.number)

    def update_c_size(self, value):
        """
        Updates graphic sizes according to the new value.

        Args:
            value (float): New graphics size
        """
        self.line_width = dp(value / 5)
        for c in range(1, self.clicks + 1):
            group = self.canvas.get_group(str(c))
            for i in group:
                if isinstance(i, Ellipse):
                    i.size = (dp(value), dp(value))
                    i.pos = (i.pos[0] + self.c_size[0] / 2 - dp(value) / 2,
                             i.pos[1] + self.c_size[1] / 2 - dp(value) / 2)
                elif isinstance(i, Line):
                    i.width = self.line_width
        if self.clicks > 0:
            self.number.font_size = dp(value) * 2
            self.number.pos = (self.number.pos[0] + self.c_size[0] - dp(value),
                               self.number.pos[1] + self.c_size[1] - dp(value))
        self.curr_line.width = self.line_width
        self.c_size = (dp(value), dp(value))

    def update_width(self, width):
        """
        Update t_width to change width for next transect made.

        Args:
            width (int): New width to use
        """
        self.t_width = width
        if len(self.points) == 1:  # Update extra width entry at start of list so avg can be taken
            self.points[0] = (self.points[0][0], self.points[0][1], width)

    def upload_mode(self, val):
        """
        Update whether in upload mode or not

        Args:
            val (bool): Whether in upload mode or not
        """
        self.uploaded = val

    def get_orthogonal(self, line):
        """
        Get a line orthogonal to line drawn by user to use as transect.

        Args:
            line: kivy.graphics.Line, line between two last clicked points

        Returns:
            4 element array of floats: Coordinates of the two endpoints of the centered
            orthogonal line with length t_width.
        """
        xyswap = False
        line = line.points

        if line[0] > line[2]:  # Always read from left point to right
            line = [line[2], line[3], line[0], line[1]]

        # Calculate line slope ensuring never dividing by zero
        if line[2] - line[0] == 0:
            m = (line[3] - line[1]) / .001
        else:
            m = (line[3] - line[1]) / (line[2] - line[0])

        # Get orthogonal slope
        if m == 0:
            m = -1 / .001
        else:
            m = -1 / m

        if abs(math.atan(m)) > (math.pi / 4):  # If angle of line is > 45 degrees will swap x and y to increase accuracy

            xyswap = True
            line = [line[1], line[0], line[3], line[2]]
            m = 1 / m

        # Find midpoint
        mid = (line[0] + (line[2] - line[0]) / 2, line[1] + (line[3] - line[1]) / 2)

        # Calculate orthogonal line points
        b = mid[1] - m * mid[0]
        xarr = np.arange(int(mid[0] - self.t_width / 2), int(mid[0] + self.t_width / 2 + 1), dtype=float)
        yarr = (xarr * m + b).tolist()
        xarr = xarr.tolist()

        # Draw points at ends of transect
        with self.canvas:
            Color(self.l_color.r, self.l_color.g, self.l_color.b)
            coords = [xarr[0], yarr[0], xarr[-1], yarr[-1]]
            if xyswap:
                coords = [yarr[0], xarr[0], yarr[-1], xarr[-1]]
            Ellipse(pos=(coords[0] - self.c_size[0] / 2, coords[1] - self.c_size[1] / 2),
                    size=self.c_size, group=str(self.clicks))
            Ellipse(pos=(coords[2] - self.c_size[0] / 2, coords[3] - self.c_size[1] / 2),
                    size=self.c_size, group=str(self.clicks))
        return coords

    def del_point(self):
        """
        Remove most recent point, line, and transect points.

        Graphics are grouped by the number of clicks made when they were created for easier deletion.
        """
        if self.clicks != 1:
            self.base.lines = self.base.lines[:-1]
        else:
            # Remove plot and width buttons from sidebar if last point of the marker
            if self.parent.dbtn in self.home.display.current:
                self.home.display.current.remove(self.parent.dbtn)
            if self.parent.width_w in self.home.display.current:
                self.home.display.current.remove(self.parent.width_w)
            self.remove_widget(self.children[0])
            # Stop drawing line between last point and cursor
            Window.unbind(mouse_pos=self.draw_line)
        self.points.remove(self.points[-1])
        self.canvas.remove_group(str(self.clicks))
        self.canvas.remove_group(str(self.clicks + 1))
        self.clicks -= 1

    def on_touch_down(self, touch):
        """
        On user click draws line, endpoints and orthogonal points.

        Args:
            touch: MouseMotionEvent, see kivy docs for details
        """
        # Draws marker line and points.
        proceed = False
        if self.uploaded:  # If being uploaded, just needs to be within image bounds
            if touch.pos[0] < self.size[0] and touch.pos[1] < self.size[1]:
                proceed = True
            else:
                self.parent.upload_fail_alert()  # Upload failed
        elif self.home.ids.view.collide_point(*self.home.ids.view.to_widget(*self.to_window(*touch.pos))):
            proceed = True  # If being clicked, must also be within viewing window

        if proceed:
            par = self.home.display.children[0].children[-2]
            self.clicks += 1
            with self.canvas:
                # Always adds point when clicked
                Color(self.l_color.r, self.l_color.g, self.l_color.b)
                Ellipse(pos=(touch.x - self.c_size[0] / 2, touch.y - self.c_size[1] / 2),
                        size=self.c_size, group=str(self.clicks))
                self.points.append((touch.x, touch.y, self.t_width))
                self.curr_line = Line(points=[], width=self.line_width, group=str(self.clicks + 1))
            # Draw line between last point and cursor whenever cursor position changes
            Window.bind(mouse_pos=self.draw_line)
            if self.clicks != 1:
                # If 2nd or more click, create a line inbetween click points
                with self.canvas:
                    Color(self.l_color.r, self.l_color.g, self.l_color.b)
                    line = Line(points=[self.points[-2][0:2], self.points[-1][0:2]],
                                width=self.line_width, group=str(self.clicks))
                # Stores orthogonal line in a SingleTransect which gets stored in base MultiTransect
                coords = self.get_orthogonal(line)
                if self.in_bounds(coords):
                    # Check if orthogonal points are within image bounds
                    x = SingleTransect(home=self.home)
                    x.line = Line(points=coords, width=self.line_width)
                    self.base.lines.append(x)
                else:
                    # Undo actions and alert user or parent
                    self.canvas.remove_group(str(self.clicks))
                    self.canvas.remove(self.curr_line)
                    self.clicks -= 1
                    self.points = self.points[:-1]
                    if self.uploaded:
                        self.parent.upload_fail_alert()
                    else:
                        functions.alert("Orthogonal point out of bounds", self.home)

            else:
                # If first click, adds marker number
                self.number = Label(text=str(len(par.children)), pos=(touch.x, touch.y), font_size=self.c_size[0] * 2)
                self.add_widget(self.number)

    def draw_line(self, instance, pos):
        """
        Draw line from most recent click point to user cursor.

        Updates anytime cursor moves. Does not draw if not current marker being drawn or if tool
        in dragging mode. Also won't draw if marker was uploaded and it was final marker.

        Args:
            instance: WindowSDL instance, current window loaded (not used by method)
            pos (tuple): 2 element tuple of floats, x and y coord of cursor position
        """
        if self.parent.children[0] == self and not self.parent.dragging:
            if self.home.ids.view.collide_point(*self.home.ids.view.to_widget(*pos)):
                mouse = self.to_widget(*pos)
                if self.size[0] >= mouse[0] >= 0 and self.size[1] >= mouse[1] >= 0:
                    with self.canvas:
                        Color(self.l_color.r, self.l_color.g, self.l_color.b)
                        self.curr_line.points = [self.points[-1][0:2], self.to_widget(pos[0], pos[1])]
        else:
            # Don't draw if not current marker or in dragging mode
            self.stop_drawing()

    def stop_drawing(self):
        """
        Remove line from most recent point to cursor.
        """
        with self.canvas:
            Color(self.l_color.r, self.l_color.g, self.l_color.b)
            self.curr_line.points = self.curr_line.points[0:2]

    def in_bounds(self, points):
        """
        Determine if points are within bounds of image.

        Args:
            points (list): List of 4 floats, X,Y coords of the two endpoints: [X1, Y1, X2, Y2]
        Returns:
            Boolean whether both endpoints are within image bounds
        """
        dots = [points[0:2], points[2:4]]
        for d in dots:
            if min(d) < 0 or any([i[0] > i[1] for i in list(zip(d, self.size))]):
                return False
        return True



"""
Marker width adjustment tool.

Creates sidebar element that allows the control of width of orthogonal transects made via the
transect marker tool.
"""

import kivy.uix as ui
from kivy.uix.textinput import TextInput
import nccut.functions as func


class MarkerWidth(ui.boxlayout.BoxLayout):
    """
    Marker width adjustment tool.

    Creates sidebar element that allows the control of width of orthogonal transects made via the
    transect marker tool.

    Attributes:
        txt: kivy.uix.textinput.TextInput widget where user enters their desired width
        btn: 'Go' Button
        marker: Current :class:`nccut.multimarker.MultiMarker` instance
        min (int): Minimum width allowed
        max (int): Maximum width allowed
    """
    def __init__(self, marker, **kwargs):
        """
        Defines graphical elements.

        Args:
            marker: Active :class:`nccut.multimarker.MultiMarker` instance
        """
        super(MarkerWidth, self).__init__(**kwargs)
        self.txt = TextInput(hint_text="Width", size_hint=(.7, 1))
        self.btn = func.RoundedButton(text="Go", size_hint=(.3, 1), font_size=marker.home.font)
        self.btn.bind(on_press=lambda x: self.update())
        self.add_widget(self.txt)
        self.add_widget(self.btn)
        self.marker = marker
        self.min = 1
        self.max = 400

    def font_adapt(self, font):
        """
        Update go button font size

        Args:
            font (float): New font size
        """
        self.btn.font_size = font

    def update(self):
        """
        Update width of current marker if given a valid width.
        """
        num = self.txt.text
        if num.isnumeric():
            num = int(num)
            if self.min <= num <= self.max:
                self.marker.update_width(num)



"""
UI and functionality for plotting and saving popup.

Creates UI elements of the popup and manages the creation of plots based on user input.
Manages the saving of the plots and user selected data.
"""

import io
import kivy.uix as ui
from kivy.graphics import Color, Rectangle
from kivy.metrics import dp
from kivy.uix.button import Button
from kivy.uix.popup import Popup
from kivy.uix.label import Label
from kivy.uix.textinput import TextInput
from kivy.uix.checkbox import CheckBox
from kivy.uix.dropdown import DropDown
import nccut.functions as func
from kivy.core.image import Image as CoreImage
import matplotlib.pyplot as plt
from PIL import Image as im
import numpy as np
import copy
import os
import pandas as pd
import json
import cv2
import img2pdf


class BackgroundDropDown(DropDown):
    """
    General class for a dropdown menu with a background color

    Attributes:
        rect: Rectangle object that serves as background to the dropdown.
    """
    def open(self, widget):
        """
        Overwrites DropDown open method to also draw a background rectangle.

        Args:
            widget: RoundedButton to which dropdown is bound
        """
        super(BackgroundDropDown, self).open(widget)
        with self.canvas.before:
            Color(rgb=[0.2, 0.2, 0.2])
            self.rect = Rectangle(size=self.size, pos=self.pos, radius=[dp(10), ])
        self.bind(pos=self.update_canvas, size=self.update_canvas)

    def update_canvas(self, *args):
        """
        Update graphics if window size is changed
        """
        self.rect.pos = self.pos
        self.rect.size = self.size


class PlotPopup(Popup):
    """
    Popup with plotting and saving selections.

    Creates UI elements of the popup, creates plots according to user selections, and saves
    plots and/or selected data.

    Attributes:
        home: Reference to root :class:`nccut.homescreen.HomeScreen` instance.
        all_transects (dict): Dictionary of data from all transects marked out by the user.
        t_type (str): 'Marker' if transects came from transect marker tool or 'Multi' if transects
            came from transect tool.
        active_transects (dict): Dictionary of currently selected transects. 'Click X', 'Click Y', and
            'Width' fields should be removed (if marker tool) to simplify plotting. Contains average
            of transects if marker tool was used with a constant transect width.
        f_type (str): If file is NetCDF file: 'NC' if only 2 dimensions, 'NC_Z' if 3 dimensions. If
            file is a JPG or PNG: 'Img'.
        config (dict): Information necessary for accessing the file. For images this is the file path and for NetCDF
            files this is a dictionary of configuration values (see
            :meth:`nccut.netcdfconfig.NetCDFConfig.check_inputs` for structure of dictionary)
        active_z (list): List of selected Z values. Empty list if 2D NetCDF or Image file.
        active_vars (list): List of selected variables. Empty list if image file.
        plot: Image containing plot.
        plotting: BoxLayout that holds plot and selection sidebar.
        title (str): Popup title
        content: BoxLayout containing all UI elements in popup
        size_hint (tuple): (width, height) of relative size of popup to window
        t_select: RoundedButton which opens transect selection dropdown menu
        v_select: RoundedButton which opens variable selection dropdown menu (Only if NetCDF file)
        z_select: RoundedButton which opens z value selection dropdown menu (Only if 3D NetCDF file)
    """
    def __init__(self, transects, home, config, **kwargs):
        """
        Defines popup UI elements and opens popup.

        Args:
            transects (dict): Dictionary of transect data from tool which opened popup
            home: Reference to root :class:`nccut.homescreen.HomeScreen` instance.
        """
        super(PlotPopup, self).__init__(**kwargs)
        self.all_transects = transects
        self.home = home
        self.f_type = list(config.keys())[0]
        self.config = config[self.f_type]
        if list(self.all_transects.keys())[0][0:-2] == "Marker":
            self.t_type = "Marker"
        else:
            self.t_type = "Multi"

        # Create nested dictionary of Booleans indicating which transects are currently selected
        self.active_transects = {}

        act = copy.deepcopy(self.all_transects)
        for key in list(act.keys()):
            self.active_transects[key] = {}
            if self.t_type == "Marker":  # Remove fields that don't get plotted
                act[key].pop("Click X")
                act[key].pop("Click Y")
                act[key].pop("Width")
            self.active_transects[key] = dict.fromkeys(act[key], False)

            # If marker and all values same width, add an average option
            if self.t_type == "Marker":
                w_lis = self.all_transects[key]['Width']
                if all(x == w_lis[0] for x in w_lis):
                    new = {"Average": False}
                    new.update(self.active_transects[key])
                    self.active_transects[key] = new

        # If not marker, start will all transects selected
        first = list(self.active_transects.keys())[0]
        self.active_transects[first] = dict.fromkeys(self.active_transects[first], True)

        # If marker start with first marker selected
        if self.t_type == "Marker":
            w_lis = self.all_transects[first]['Width']
            if all(x == w_lis[0] for x in w_lis):
                self.active_transects[first]["Average"] = False

        # Initialize dropdown selections
        if self.f_type == "netcdf":
            if self.config['z'] == "N/A":
                self.active_z = []
            else:
                self.active_z = [self.config['z_val']]
            self.active_vars = [self.config["var"]]
        else:
            self.active_vars = []
            self.active_z = []

        # Get plot for initial selections
        self.active_data = self.get_data()
        self.plot_active()
        temp = io.BytesIO()
        plt.savefig(temp, format="png")
        temp.seek(0)
        plt.close()

        # Popup Graphics Code

        self.plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture,
                                   size_hint=(0.7, 1), fit_mode="contain")
        # Plot
        self.plotting = ui.boxlayout.BoxLayout(spacing=dp(20), size_hint=(1, 0.9))
        self.plotting.add_widget(self.plot)
        self.title = "Plot Transects"
        self.content = ui.boxlayout.BoxLayout(orientation='vertical', spacing=dp(20), padding=dp(20))
        self.size_hint = (0.8, 0.8)
        sidebar = ui.boxlayout.BoxLayout(orientation='vertical', size_hint=(0.3, 1), padding=dp(10), spacing=dp(20))

        # Saving Data/Plotting buttons

        buttons = ui.boxlayout.BoxLayout(orientation='horizontal', size_hint=(1, .1), spacing=dp(10))

        f_m = 0.8
        a_data_btn = func.RoundedButton(text="Save All Data", size_hint=(.15, 1), font_size=self.home.font * f_m)
        a_data_btn.bind(on_press=lambda x: self.file_input('a_data'))

        s_data_btn = func.RoundedButton(text="Save Selected Data", size_hint=(.15, 1), font_size=self.home.font * f_m)
        s_data_btn.bind(on_press=lambda x: self.file_input('s_data'))

        png_btn = func.RoundedButton(text='Save Plot to PNG', size_hint=(.15, 1), font_size=self.home.font * f_m)
        png_btn.bind(on_press=lambda x: self.file_input('png'))

        pdf_btn = func.RoundedButton(text='Save Plot to PDF', size_hint=(.15, 1), font_size=self.home.font * f_m)
        pdf_btn.bind(on_press=lambda x: self.file_input('pdf'))

        buttons.add_widget(a_data_btn)
        buttons.add_widget(s_data_btn)
        buttons.add_widget(png_btn)
        buttons.add_widget(pdf_btn)

        # Transect selection
        t_box = ui.boxlayout.BoxLayout(size_hint=(1, 0.2), spacing=dp(30))
        lab = Label(text="Select Transects: ", size_hint=(0.3, 1), font_size=self.home.font)
        t_box.add_widget(lab)
        self.t_select = func.RoundedButton(text="Select...", size_hint=(0.7, 1), font_size=self.home.font)
        t_box.add_widget(self.t_select)
        if self.t_type == "Marker":
            t_drop = self.get_marker_dropdown()
        else:
            t_drop = self.get_cut_dropdown('Multi')
        self.t_select.bind(on_press=lambda x: t_drop.open(self.t_select))
        sidebar.add_widget(t_box)

        if self.f_type == "netcdf":
            # Variable Selection
            v_box = ui.boxlayout.BoxLayout(size_hint=(1, 0.2), spacing=dp(30))
            v_box.add_widget(Label(text="Select Variables: ", size_hint=(0.3, 1), font_size=self.home.font))
            self.v_select = func.RoundedButton(text="Select...", size_hint=(0.7, 1),
                                               font_size=self.home.font)
            v_box.add_widget(self.v_select)
            v_drop = self.get_var_dropdown()
            self.v_select.bind(on_press=lambda x: v_drop.open(self.v_select))
            sidebar.add_widget(v_box)
            if self.config['z'] != "N/A":
                # Z Selection
                z_box = ui.boxlayout.BoxLayout(size_hint=(1, 0.2), spacing=dp(30))
                z_box.add_widget(Label(text="Select Z Values: ", size_hint=(0.3, 1), font_size=self.home.font))
                self.z_select = func.RoundedButton(text="Select...", size_hint=(0.7, 1),
                                                   font_size=self.home.font)
                z_box.add_widget(self.z_select)
                z_drop = self.get_z_dropdown()
                self.z_select.bind(on_press=lambda x: z_drop.open(self.z_select))
                sidebar.add_widget(z_box)

                # All Z option
                zp_box = ui.boxlayout.BoxLayout(size_hint=(1, 0.2), spacing=dp(30))
                zp_btn = func.RoundedButton(text="Plot all Z as Img", font_size=self.home.font)
                zp_btn.bind(on_press=lambda x: self.get_all_z_plot())
                zp_box.add_widget(zp_btn)
                sidebar.add_widget(zp_box)

                allz_btn = func.RoundedButton(text="Save All Z Data", size_hint=(.2, 1), font_size=self.home.font * f_m)
                allz_btn.bind(on_press=lambda x: self.file_input('all_z'))
                buttons.add_widget(allz_btn)
            else:
                # Spacer if 2D NetCDF
                sidebar.add_widget(Label(text="", size_hint=(1, 0.6)))
        else:
            # Spacer if Image
            sidebar.add_widget(Label(text="", size_hint=(1, 0.8)))

        self.plotting.add_widget(sidebar)
        self.content.add_widget(self.plotting)

        close = func.RoundedButton(text="Close", size_hint=(.2, 1))
        close.bind(on_press=self.dismiss)
        buttons.add_widget(close)
        self.content.add_widget(buttons)
        self.open()

    def file_input(self, type):
        """
        Popup window for user to give name for plot/json file to be saved.

        Args:
            type (str): String corresponding to what is being saved:
                'data': Transect data for selections
                'all_z_data': Transect data for selections for all z dimension values
                'png': Current plot as a PNG file
                'pdf': Current plot as a PDF file
        """
        content = ui.boxlayout.BoxLayout(orientation='horizontal')
        popup = Popup(title="File Name", content=content, size_hint=(0.5, 0.15))
        txt = TextInput(size_hint=(0.7, 1), hint_text="Enter File Name")
        content.add_widget(txt)
        go = Button(text="Ok", size_hint=(0.1, 1))
        if type == "s_data":
            go.bind(on_press=lambda x: self.download_selected_data(txt.text))
        elif type == "a_data":
            go.bind(on_press=lambda x: self.download_all_data(txt.text))
        elif type == "all_z":
            go.bind(on_press=lambda x: self.download_all_z_data(txt.text))
        elif type == "png":
            go.bind(on_press=lambda x: self.download_png_plot(txt.text))
        elif type == "pdf":
            go.bind(on_press=lambda x: self.download_pdf_plot(txt.text))
        go.bind(on_release=lambda x: self.close_popups(popup))
        close = Button(text="Close", size_hint=(0.2, 1), font_size=self.home.font)
        close.bind(on_press=popup.dismiss)
        content.add_widget(go)
        content.add_widget(close)
        popup.open()

    def close_popups(self, fpop):
        """
        Close file name popup and plot popup

        Args:
            fpop: reference to file name popup
        """
        fpop.dismiss()
        self.dismiss()

    def download_png_plot(self, f_name):
        """
        Download the current plot as a PNG file if valid file name given

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".png")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:
            path = self.home.rel_path / (file + ".png")
            self.plot.texture.save(str(path.absolute()))
            img = cv2.flip(cv2.imread(str(path.absolute())), 0)
            cv2.imwrite(str(path.absolute()), img)
            func.alert("Download Complete", self.home)

    def download_pdf_plot(self, f_name):
        """
        Download the current plot as a PDF file if valid file name given

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".pdf")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:
            ipath = self.home.rel_path / (file + ".png")
            ppath = self.home.rel_path / (file + ".pdf")
            self.plot.texture.save(str(ipath.absolute()))
            img = cv2.flip(cv2.imread(str(ipath.absolute())), 0)
            cv2.imwrite(str(ipath.absolute()), img)
            with open(str(ppath.absolute()), "wb") as f:
                f.write(img2pdf.convert(str(ipath.absolute())))
            os.remove(ipath)
            func.alert("Download Complete", self.home)

    def download_selected_data(self, f_name):
        """
        Downloads selected transect data into a JSON file if valid file name given.

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".json")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:  # Build correctly formatted JSON file
            dat = copy.copy(self.active_data)
            if len(self.active_vars) == 0:  # If Image
                final = self.add_marker_info(dat)
            elif len(self.active_z) == 0:  # If 2D NetCDF
                final = {}
                for var in list(dat.keys()):
                    final[var] = self.add_marker_info(dat[var])
            else:  # If 3D NetCDF
                final = {}
                for var in list(dat.keys()):
                    final[var] = {}
                    for z in list(dat[var].keys()):
                        final[var][z] = self.add_marker_info(dat[var][z])
            with open(self.home.rel_path / (file + ".json"), "w") as f:
                json.dump(final, f)

            func.alert("Download Complete", self.home)

    def download_all_data(self, f_name):
        """
        Downloads selected data for all transects/markers into a JSON file if valid file name given.

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".json")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:  # Build JSON file
            original = copy.copy(self.active_transects)
            for m in list(self.active_transects.keys()):
                for t in list(self.active_transects[m].keys()):
                    self.active_transects[m][t] = True
            self.active_data = self.get_data()
            self.download_selected_data(f_name)
            self.active_transects = original
            self.active_data = self.get_data()

    def add_marker_info(self, dicti):
        """
        Adds back fields removed for plotting purposes if marker tool was used

        Args:
            dicti (dict): Dictionary of transect data from either transect marker or regular transect tool

        Returns:
            Dictionary of transect data with 'Click X', 'Click Y', and 'Width' fields added back
                in if transect marker tool was used.
        """
        if list(dicti.keys())[0] != "Multi":
            for marker in list(dicti.keys()):
                dicti[marker]['Click X'] = self.all_transects[marker]['Click X']
                dicti[marker]['Click Y'] = self.all_transects[marker]['Click Y']
                dicti[marker]['Width'] = self.all_transects[marker]['Width']
        return dicti

    def download_all_z_data(self, f_name):
        """
        Get and download data for all selected variables for all z dimension values.

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".json")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:
            original = copy.copy(self.active_z)
            self.active_z = [str(z) for z in self.config['file'].coords[self.config['z']].data]
            self.active_data = self.get_data()
            self.download_selected_data(f_name)
            self.active_z = original
            self.active_data = self.get_data()

    def download_all_z_data2(self, f_name):
        """
        Get and download data for all selected variables for all z dimension values.

        Args:
            f_name (str): Proposed file name
        """
        file = func.check_file(self.home.rel_path, f_name, ".json")
        if file is False:
            func.alert("Invalid File Name", self.home)
            return
        else:
            dat = {}
            ds = self.config['file']
            z_arr = [str(z) for z in ds.coords[self.config['z']].data]
            for var in self.active_vars:
                config = copy.copy(self.config)
                config['var'] = var
                ds = ds[config['var']]
                dat[var] = {}
                ds.load()
                ds = ds.rename({config['y']: "y", config['x']: "x", config['z']: "z"})
                ds = ds.transpose('y', 'x', 'z')
                ds['z'] = ds['z'].astype(str)
                ds = np.flip(ds.data, 0)
                for d, d_lab in enumerate(z_arr):
                    dat[var][str(d_lab)] = {}
                    curr = ds[:, :, d]
                    for key in list(self.active_transects.keys()):
                        # All markers selected (if marker tool used)
                        dat[var][str(d_lab)][key] = {}
                        for cut in list(self.active_transects[key].keys()):
                            # All selected transects
                            if self.active_transects[key][cut]:
                                if cut == "Average":
                                    dat[var][str(d_lab)][key][cut] = self.get_average(key, curr)
                                else:
                                    dat[var][str(d_lab)][key][cut] = func.ip_get_points(self.all_transects[key][cut],
                                                                                        curr,
                                                                                        self.f_type == "netcdf")
                        if len(dat[var][str(d_lab)][key]) == 0:
                            dat[var][str(d_lab)].pop(key)
                    dat[var][str(d_lab)] = self.add_marker_info(dat[var][str(d_lab)])

            with open(self.home.rel_path / (file + ".json"), "w") as f:
                json.dump(dat, f)
            func.alert("Download Complete", self.home)

    def get_marker_dropdown(self):
        """
        Build dropdown menu for marker options with sub-menus for the individual transects

        Returns:
            :class:`nccut.plotpopup.BackgroundDropDown` for marker options
        """
        # Get dropdown for marker options
        marker_list = BackgroundDropDown(auto_width=False, width=dp(180), max_height=dp(300))
        for i in list(self.all_transects.keys()):
            m_box = ui.boxlayout.BoxLayout(spacing=dp(10), padding=dp(10), size_hint_y=None, height=dp(40),
                                           width=dp(180))
            btn = func.RoundedButton(text=i, size_hint=(0.5, 1))
            btn.bind(on_press=lambda but=btn, txt=i: self.cut_drop(txt, but))
            m_box.add_widget(btn)
            marker_list.add_widget(m_box)
        return marker_list

    def cut_drop(self, marker, button):
        """
        Attaches transect dropdowns to marker buttons in marker dropdown menu

        Args:
            marker (str): Marker label 'Marker #'
            button: RoundedButton, marker's button in marker dropdown menu
        """
        temp_cut_drop = self.get_cut_dropdown(marker)
        temp_cut_drop.open(button)

    def get_cut_dropdown(self, key):
        """
        Build dropdown menu for selecting transects.

        Args:
            key (str): Name of marker selecting from 'Marker #' or 'Multi' if marker tool wasn't used

        Returns:
            :class:`nccut.plotpopup.BackgroundDropDown` for transect options
        """
        # Get dropdown for transect options
        drop = BackgroundDropDown(auto_width=False, width=dp(180), max_height=dp(200))
        all_box = ui.boxlayout.BoxLayout(spacing=dp(10), padding=dp(10), size_hint_y=None, height=dp(40),
                                         width=dp(180))
        drop.add_widget(all_box)
        all_btn = func.RoundedButton(text="Select All", size_hint=(0.5, 1))
        for i in list(self.active_transects[key].keys()):
            c_box = ui.boxlayout.BoxLayout(spacing=dp(5), size_hint_y=None, height=dp(40), width=dp(180))
            lab = Label(text=i, size_hint=(0.5, 1))
            c_box.add_widget(lab)
            check = CheckBox(active=self.active_transects[key][i], size_hint=(0.5, 1))
            check.bind(active=lambda x, y, m=key, t=i: self.on_transect_checkbox(x, m, t))
            c_box.add_widget(check)
            drop.add_widget(c_box)
        all_btn.bind(on_press=lambda x: self.select_all(drop.children[0].children[:-1]))
        all_box.add_widget(all_btn)
        return drop

    def select_all(self, boxes):
        """
        If all checkboxes are checked, uncheck all boxes. Otherwise check all boxes. If a box is the only box checked
        across all markers it will remain checked no matter what.

        Args:
            boxes: List of BoxLayouts containing checkbox widgets
        """
        all_true = True
        for c_box in boxes:
            if not c_box.children[0].active:
                c_box.children[0].active = not c_box.children[0].active
                all_true = False
        if all_true:
            for c_box in boxes:
                c_box.children[0].active = False

    def on_transect_checkbox(self, check, marker, cut, *args):
        """
        Updates plot when a transect checkbox is clicked. Safeguards so at least one transect
        is always selected.

        Args:
            check: Reference to kivy.uix.checkbox.CheckBox in transect list
            marker (str): Name of marker selecting from 'Marker #' or 'Multi' if marker tool wasn't used
            cut (str): Name of transect 'Cut #'
        """
        # Select or deselect transect
        self.active_transects[marker][cut] = not self.active_transects[marker][cut]

        # Check this isn't the last transect selected
        count = 0
        for key in list(self.active_transects.keys()):  # Count current transects
            count += sum(self.active_transects[key].values())
        if count == 0:  # If last transect unchecked, recheck and ignore
            self.active_transects[marker][cut] = not self.active_transects[marker][cut]
            check.active = True
            return
        else:
            # Update current data
            self.active_data = self.get_data()
            # Update plot
            self.plotting.remove_widget(self.plot)
            self.plot_active()
            temp = io.BytesIO()
            plt.savefig(temp, format="png")
            temp.seek(0)
            plt.close()
            self.plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture,
                                       size_hint=(0.7, 1), fit_mode="contain")
            self.plotting.add_widget(self.plot, len(self.plotting.children))

    def get_var_dropdown(self):
        """
        Get dropdown for NetCDF variable options. Variable only available if dimensions match current variable

        Returns:
            :class:`nccut.plotpopup.BackgroundDropdown` menu of variable options
        """
        file = self.config['file']
        var_list = BackgroundDropDown(auto_width=False, width=dp(180), max_height=dp(300))
        for var in list(file.keys()):
            if file[self.config['var']].dims == file[var].dims:  # Dimensions must match variable in viewer
                v_box = ui.boxlayout.BoxLayout(spacing=dp(10), padding=dp(10), size_hint_y=None, height=dp(40),
                                               width=dp(180))
                lab = Label(text=var, size_hint=(0.5, 1))
                v_box.add_widget(lab)
                check = CheckBox(active=var in self.active_vars, size_hint=(0.5, 1))
                check.bind(active=lambda x, y, var=var: self.on_var_checkbox(x, var))
                v_box.add_widget(check)
                var_list.add_widget(v_box)
        return var_list

    def on_var_checkbox(self, check, var, *args):
        """
        Updates plot when a variable checkbox is clicked. Safeguards so at least one variable
        is always selected.

        Args:
            check: Reference to kivy.uix.checkbox.CheckBox in variable list
            var (str): String, name of variable
        """
        if var in self.active_vars:
            self.active_vars.remove(var)
        else:
            self.active_vars.append(var)

        if not self.active_vars:  # If last variable add variable back
            self.active_vars.append(var)
            check.active = True
            return
        self.active_data = self.get_data()

        # Update plot
        self.plotting.remove_widget(self.plot)
        self.plot_active()
        temp = io.BytesIO()
        plt.savefig(temp, format="png")
        temp.seek(0)
        plt.close()
        self.plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture,
                                   size_hint=(0.7, 1), fit_mode="contain")
        self.plotting.add_widget(self.plot, len(self.plotting.children))

    def get_z_dropdown(self):
        """
        Get dropdown for NetCDF z value selections

        Returns:
            :class:`nccut.plotpopup.BackgroundDropDown` menu of z value options
        """
        z_list = BackgroundDropDown(auto_width=False, width=dp(180), max_height=dp(300))
        for z in list(self.config['file'].coords[self.config['z']].data):
            z_box = ui.boxlayout.BoxLayout(spacing=dp(10), padding=dp(10), size_hint_y=None, height=dp(40),
                                           width=dp(180))
            lab = Label(text=str(z), size_hint=(0.5, 1))
            z_box.add_widget(lab)
            check = CheckBox(active=str(z) in self.active_z, size_hint=(0.5, 1))
            check.bind(active=lambda x, y, z=str(z): self.on_z_checkbox(x, z))
            z_box.add_widget(check)
            z_list.add_widget(z_box)
        return z_list

    def on_z_checkbox(self, check, z, *args):
        """
        Updates plot when a z value checkbox is clicked. Safeguards so at least one z value
        is always selected.

        Args:
            check: Reference to kivy.uix.checkbox.CheckBox in variable list
            z (str): Name of z value
            *args: Unused args passed to method
        """
        if z in self.active_z:
            self.active_z.remove(z)
        else:
            self.active_z.append(z)

        if not self.active_z:  # If last z value add z value back
            self.active_z.append(z)
            check.active = True
            return
        self.active_data = self.get_data()

        # Update plot
        self.plotting.remove_widget(self.plot)
        self.plot_active()
        temp = io.BytesIO()
        plt.savefig(temp, format="png")
        temp.seek(0)
        plt.close()
        self.plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture,
                                   size_hint=(0.7, 1), fit_mode="contain")
        self.plotting.add_widget(self.plot, len(self.plotting.children))

    def get_all_z_plot(self):
        """
        Determines if settings are okay to do an all z value plot. If so calls for plot, if not creates an error
        popup.
        """
        count = 0
        for key in list(self.active_transects.keys()):
            # Count current transects selected
            count += sum(self.active_transects[key].values())
        if count == 1:
            # Go ahead and plot
            self.plot_all_z()
            self.plotting.remove_widget(self.plot)
            temp = io.BytesIO()
            plt.savefig(temp, format="png")
            temp.seek(0)
            plt.close()
            self.plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture,
                                       size_hint=(0.7, 1), fit_mode="contain")
            self.plotting.add_widget(self.plot, len(self.plotting.children))
        else:
            # Error popup if more than one transect is selected
            content = ui.boxlayout.BoxLayout()
            error = Popup(title="Error", content=content, size_hint=(0.5, 0.15))
            lab = Label(text="Please only select one transect", size_hint=(0.8, 1))
            close = Button(text="Close", size_hint=(0.2, 1))
            close.bind(on_press=error.dismiss)
            content.add_widget(lab)
            content.add_widget(close)
            error.open()

    def plot_all_z(self):
        """
         Gather all z value plots for all selected variables into one figure

         Returns:
            Final completed figure
        """
        # Determine subplot layout
        num = len(self.active_vars)
        if num <= 1:
            col = 1
            row = 1
        else:
            col = 2
            if num % 2 == 0:
                row = int(num / 2)
            else:
                row = int((num + 1) / 2)
        fig, ax = plt.subplots(row, col)

        count = 0
        for var in self.active_vars:
            # Plot variable and assign to subplot
            count += 1
            if count % 2 == 0:
                c = 1
                r = int((count / 2) - 1)
            else:
                c = 0
                r = int(((count + 1) / 2) - 1)
            if row == 1 and col == 1:
                pos = self.z_ip(var, ax)
                fig.colorbar(pos, ax=ax)
            elif row == 1:
                pos = self.z_ip(var, ax[c])
                fig.colorbar(pos, ax=ax[c])
            else:
                pos = self.z_ip(var, ax[r, c])
                fig.colorbar(pos, ax=ax[r, c])
        if len(self.active_vars) % 2 == 1 and len(self.active_vars) > 1:
            plt.delaxes(ax[r, 1])
        return fig

    def z_ip(self, var, ax):
        """
        Gather data and plot all z values for given variable

        Args:
            var (str): Variable name
            ax: Plot axis on which to make plot

        Returns:
            Axis with completed plot
        """
        # Get transect coordinates
        v = self.active_data[var]
        z = v[next(iter(v))]
        marker = next(iter(z))
        tran = next(iter(z[marker]))
        points = self.all_transects[marker][tran]
        width = len(z[marker][tran]['x'])

        ds = self.config['file']
        z_len = len(ds.coords[self.config['z']].data)
        config = copy.copy(self.config)
        config['var'] = var
        ds = ds[config['var']]

        # Load entire dataset (if big file this will take a while but decreases total plot time)
        ds.load()
        ds = ds.rename({config['y']: "y", config['x']: "x", config['z']: "z"})
        ds = ds.transpose('y', 'x', 'z')
        ds['z'] = ds['z'].astype(str)
        ds = np.flip(ds.data, 0)

        # Array of data values at x, y pairs for each z
        all_z = np.empty(shape=(z_len, width))
        c = 0

        for d in range(0, z_len):
            # Get transect data for each z level and add to array
            curr = ds[:, :, d]
            dat = func.ip_get_points(points, curr, self.f_type == "netcdf")
            all_z[c, :] = dat['Cut']
            c += 1

        # Plot array
        pos = ax.imshow(all_z)
        ax.set_ylabel(self.config['z'])
        ax.set_xlabel("Along Transect Point")
        ax.set_title(var)
        return pos

    def plot_active(self):
        """
        Creates figure of plots of currently selected data

        Returns:
            Final completed figure
        """
        # Determine subplot layout based on the number of active variables selected
        num = len(self.active_vars)
        if num <= 1:
            col = 1
            row = 1
        else:
            col = 2
            if num % 2 == 0:
                row = int(num / 2)
            else:
                row = int((num + 1) / 2)
        fig, ax = plt.subplots(row, col)

        if self.f_type == "image":  # If image just plot
            names = self.plot_single(self.active_data, ax, "Mean RGB Value")
        else:
            count = 0
            for var in self.active_vars:
                # Subplot for each variable
                count += 1
                if count % 2 == 0:
                    c = 1
                    r = int((count / 2) - 1)
                else:
                    c = 0
                    r = int(((count + 1) / 2) - 1)
                if row == 1 and col == 1:
                    names = self.plot_single(self.active_data[var], ax, var)
                elif row == 1:
                    names = self.plot_single(self.active_data[var], ax[c], var)
                else:
                    names = self.plot_single(self.active_data[var], ax[r, c], var)
            if len(self.active_vars) % 2 == 1 and len(self.active_vars) > 1:
                # If unused subplot in layout, delete it
                plt.delaxes(ax[r, 1])
        fig.legend(names, title="Legend", bbox_to_anchor=(1, 1))
        return fig

    def plot_single(self, data, ax, label):
        """
        Creates a single plot of selected transect data for a single variable (if NetCDF)

        Args:
            data: Currently selected transect data. Either self.active_data itself or a sub-dictionary
            ax: Plot axis on which to make plot
            label: Label for the y axis of plot. If NetCDF use variable if Image use 'Mean RGB Value'

        Returns:
            List of string labels for the plot's legend
        """
        # Create a non-nested dictionary of selections with appropriate labels
        dat = copy.copy(data)
        plot_dat = {}

        if list(dat.keys())[0][0:6] != "Marker" and list(dat.keys())[0] != "Multi":
            # Gather data for all transects selected across all markers for all Z levels selected
            for z in list(dat.keys()):
                if len(z) >= 12:
                    z_name = z[:12] + "..."
                else:
                    z_name = z
                for obj in list(dat[z].keys()):
                    if obj == "Multi":
                        title = "Z: " + z_name + " "
                    else:
                        title = "Z: " + z_name + " M" + obj[-1] + " "
                    for cut in list(dat[z][obj].keys()):
                        if cut == "Average":
                            plot_dat[title + cut] = dat[z][obj][cut]
                        else:
                            plot_dat[title + cut] = dat[z][obj][cut]["Cut"]
        else:
            for obj in list(dat.keys()):
                # Gather data for all transects selected across all markers
                if obj == "Multi":
                    title = ""
                else:
                    title = "M" + obj[-1] + " "
                for cut in list(dat[obj].keys()):
                    if cut == "Average":
                        plot_dat[title + cut] = dat[obj][cut]
                    else:
                        plot_dat[title + cut] = dat[obj][cut]["Cut"]

        # Plot data by turning dictionary into a data frame
        df = pd.DataFrame.from_dict(dict([(k, pd.Series(v)) for k, v in plot_dat.items()]))
        x = np.asarray(df.index)
        axis = (x - x[0]) / (x[-1] - x[0])
        ax.plot(axis, df)

        ax.set_ylabel(label.capitalize())
        if not self.f_type == "netcdf":
            ax.set_ylim(ymin=0)
        ax.set_xlabel("Normalized Long Transect Distance")
        plt.tight_layout()
        # Return dataframe column names for legend
        return df.columns

    def get_data(self):
        """
        Gathers transect data for currently selected variables, z values, and transects.

        Returns:
            Nested dictionary of transect data with hierarchy:
                Variables -> Z values -> Markers/Multi -> Transects -> X,Y, Cut
            If only 2D NetCDF file there is no Z values level. If an image there is no Variables or Z Values level.
        """
        # Get transect data for list of active transect points
        config = copy.copy(self.config)
        values = {}
        if len(self.active_vars) >= 1:
            # If data has variables
            for var in self.active_vars:
                values[var] = {}
                config["var"] = var
                if len(self.active_z) >= 1:
                    # If data has multiple Z levels
                    for z in self.active_z:
                        # Multi Z selection
                        values[var][z] = {}
                        config["z_val"] = z
                        curr = func.sel_data(config)
                        for key in list(self.active_transects.keys()):
                            # All markers selected
                            values[var][z][key] = {}
                            for cut in list(self.active_transects[key].keys()):
                                # All selected transects
                                if self.active_transects[key][cut]:
                                    if cut == "Average":
                                        values[var][z][key][cut] = self.get_average(key, curr)
                                    else:
                                        values[var][z][key][cut] = func.ip_get_points(self.all_transects[key][cut],
                                                                                      curr, self.f_type == "netcdf")
                            if len(values[var][z][key]) == 0:
                                values[var][z].pop(key)
                else:
                    curr = func.sel_data(config)
                    # Single Z Selection
                    for key in list(self.active_transects.keys()):
                        # All markers selected
                        values[var][key] = {}
                        for cut in list(self.active_transects[key].keys()):
                            if self.active_transects[key][cut]:
                                # All transects selected
                                if cut == "Average":
                                    values[var][key][cut] = self.get_average(key, curr)
                                else:
                                    values[var][key][cut] = func.ip_get_points(self.all_transects[key][cut],
                                                                               curr, self.f_type == "netcdf")
                        if len(values[var][key]) == 0:
                            values[var].pop(key)
        else:
            curr = im.open(config)
            # Image data (no variables, no Z levels)
            for key in list(self.active_transects.keys()):
                # All markers selected
                values[key] = {}
                for cut in list(self.active_transects[key].keys()):
                    # All transects selected
                    if self.active_transects[key][cut]:
                        if cut == "Average":
                            values[key][cut] = self.get_average(key, curr)
                        else:
                            values[key][cut] = func.ip_get_points(self.all_transects[key][cut], curr,
                                                                  self.f_type == "netcdf")
                if len(values[key]) == 0:
                    values.pop(key)
        return values

    def get_average(self, key, curr):
        """
        Finds average of all transects in a marker. Transect width must be the same for the entire marker.

        Args:
            key (str): 'Marker #'
            curr: 2D array, currently loaded dataset

        Returns:
            1D array of length of transect width containing average transect values for the marker.
        """
        dat = np.zeros(self.all_transects[key]['Width'][0])
        for cut in list(self.all_transects[key].keys())[3:]:
            dat += func.ip_get_points(self.all_transects[key][cut], curr, self.f_type == "netcdf")['Cut']
        dat = dat / len(list(self.all_transects[key].keys())[3:])
        return list(dat)



"""
Builds app and sets initial window.

Creates the widget tree and sets the initial window size. To load app, run ``NcCut().run()``

"""

from progress.bar import ChargingBar
bar = ChargingBar("Loading App", max=3)
import os
import re
import logging
import copy
from nccut.logger import get_logging_level
_LOG_LEVEL_ = copy.copy(get_logging_level())
os.environ["KIVY_NO_ARGS"] = "true"
os.environ["KCFG_KIVY_LOG_LEVEL"] = _LOG_LEVEL_.lower()
bar.next()
from kivy.metrics import dp
import kivy
import kivy.uix as ui
from kivy.app import App
import platform
import argparse
logging.getLogger().setLevel(getattr(logging, _LOG_LEVEL_, None))
bar.next()
from nccut.homescreen import HomeScreen


class NcCut(App):
    """
    Builds app and widget tree.

    Creates the initial window and ensures font sizes in the app update uniformly
    when the window resizes.

    """
    def __init__(self, file=None, **kwargs):
        """
        Initialize app with file if included via command line
        """
        super(NcCut, self).__init__(**kwargs)
        self.file = file

    def on_start(self):
        """
        Sets initial window size according to operating system.
        """

        # Set logger level to suppress or allow dependency debug messages
        logging.getLogger().setLevel(getattr(logging, get_logging_level().upper(), None))
        # Kivy has a mobile app emulator that needs to be turned off for computer app
        kivy.config.Config.set('input', 'mouse', 'mouse,disable_multitouch')
        kivy.config.Config.set('kivy', 'exit_on_escape', '0')
        win = kivy.core.window.Window
        logging.getLogger("kivy").setLevel(logging.ERROR)
        if platform.system() == "Darwin":  # macOS
            win.size = (dp(500), dp(300))
            win.minimum_width, win.minimum_height = (dp(400), dp(225))
        else:
            win.size = (dp(750), dp(450))
            win.minimum_width, win.minimum_height = (dp(600), dp(350))
        logging.getLogger("kivy").setLevel(_LOG_LEVEL_)

        win.bind(on_resize=self.on_resize)

    def on_resize(self, *args):
        """
        Triggers font adjustments when the window size is adjusted.

        Args:
            *args: Accepts args App class supplies though they aren't needed.
        """
        self.root.get_screen("HomeScreen").font_adapt()

    def build(self):
        """
        Override App class build method and return widget tree.

        Returns
            Root of widget tree
        """
        root = ui.screenmanager.ScreenManager()
        home = HomeScreen(name="HomeScreen", file=self.file)
        root.add_widget(home)
        return root


def run():
    """
    Runs app with command line file entry
    """
    bar.next()
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs='?', default=None, help="File path for image or NetCDF file")
    args = parser.parse_args()
    file = args.file
    if not file:
        bar.next()
        bar.finish()
        NcCut().run()
    elif not os.path.isfile(file):
        print("ERROR: File Not Found")
    elif len(re.findall(r'[^A-Za-z0-9_:\\.\-/]', str(file))) > 0:
        print("ERROR: Invalid File Name")
    elif not os.path.splitext(file)[1] in [".jpg", ".jpeg", ".png", ".nc"]:
        print("ERROR: File not an Image or NetCDF File")
    else:
        bar.next()
        bar.finish()
        NcCut(file).run()



"""
Functionality for viewing window and interactive display.

This module manages the scrolling, rotation, and flipping of the image. It also executes the creation of
and updates made to the image/dataset being displayed. The dragability of the image is managed by the parent
ScatterLayout class. Manages the creation and deletion of tools.
"""

import kivy
from kivy.graphics.transformation import Matrix
from kivy.uix.scatterlayout import ScatterLayout
import kivy.uix as ui
from kivy.core.image import Image as CoreImage
from PIL import Image as im
import numpy as np
import cv2
import copy
import io
import nccut.functions as func
from nccut.multitransect import MultiTransect
from nccut.multimarker import MultiMarker
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


class FileDisplay(ScatterLayout):
    """
    Creation and management of the interactive display.

    If file is a NetCDF file it turns dataset into an image according to settings. Loads image and updates it when
    settings are changed. Also manages the scrolling, flipping, and rotating of the image.

    Attributes:
        config (dict): A dictionary holding info about the file necessary for loading, updating, and accessing data from
            the file. Highest level should have one key that is the name of the file type ("image" or "netcdf") whose
            value is the necessary configuration settings. For images, the config dictionary has form
            {"image": str(file_path)}. For a netcdf file the value is a dictionary of configuration values (see
            :meth:`nccut.netcdfconfig.NetCDFConfig.check_inputs` for structure of dictionary)
        f_type (str): File type being loaded ("image" or "netcdf")
        home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        pos (tuple): Position of viewer. Used to properly place transect widgets on screen.
        sidebar (list): Reference to list of sidebar buttons
        og_sidebar (list): Original state of sidebar before any tools added widgets
        im: kivy.core.CoreImage made from data array
        byte: io.BytesIO object containing image made from NetCDF dataset loaded in memory
        img: kivy.uix.image.Image UI element which displays the image over the scatter object
        tool: Reference to currently loaded tool
        nc_data: If file is a NetCDF file, actively loaded data array
        dragging (bool): Whether in dragging mode or not
        editing (bool): Whether in editing mode or not
        t_mode (bool): Whether a tool is currently loaded
        contrast (int): Int from -127 to 127, contrast value to use when making image from NetCDF file
        l_col (str): Line color for transect tools: 'Blue', 'Green' or 'Orange'
        cir_size (float): Circle size for transect tools
        cmaps (dict): Dictionary of colormap names mapping to colormap values from cv2
        colormap: current colormap data
        action_lbl: Actions label to be loaded when in transect mode
        drag_btn: Drag button to be loaded when in transect mode
        edit_btn: Edit button to be loaded when in transect mode
        back_btn: Back button for editing mode
        delete_line_btn: Delete line button
        delete_point_btn: Delete point button
        edit_widgets (list): List of widgets that must be added to screen when entering editing mode
        current (list): List to hold other widgets when they are replaced by editing mode widgets
    """
    def __init__(self, home, f_config, **kwargs):
        """
        Initializes settings and defines editing mode buttons.

        Args:
            home: Reference to root :class:`nccut.homescreen.HomeScreen` reference
            f_config (dict): A dictionary holding info about the file necessary for loading, updating, and accessing
                data from the file. Highest level should have one key that is the name of the file type ("image" or
                "netcdf") whose value is the necessary configuration settings. For images, the config dictionary has
                form {"image": str(file)}. For a netcdf file the value is a dictionary of configuration values (see
                :meth:`nccut.netcdfconfig.NetCDFConfig.check_inputs` for structure of dictionary)
        """
        super(FileDisplay, self).__init__(**kwargs)

        self.config = f_config
        self.f_type = list(f_config.keys())[0]
        self.home = home
        self.pos = self.home.ids.view.pos
        self.sidebar = self.home.ids.sidebar
        self.og_sidebar = copy.copy(self.home.ids.sidebar.children)

        self.im = None
        self.byte = None
        self.img = None
        self.tool = None
        self.nc_data = None

        self.dragging = False
        self.editing = False
        self.t_mode = False

        self.contrast = 1.0
        self.l_col = "Blue"
        self.cir_size = 10

        self.cmaps = {"Autumn": cv2.COLORMAP_AUTUMN, "Bone": cv2.COLORMAP_BONE, "Jet": cv2.COLORMAP_JET,
                      "Winter": cv2.COLORMAP_WINTER, "Rainbow": cv2.COLORMAP_RAINBOW, "Ocean": cv2.COLORMAP_OCEAN,
                      "Summer": cv2.COLORMAP_SUMMER, "Spring": cv2.COLORMAP_SPRING, "Cool": cv2.COLORMAP_COOL,
                      "HSV": cv2.COLORMAP_HSV, "Pink": cv2.COLORMAP_PINK, "Hot": cv2.COLORMAP_HOT,
                      "Parula": cv2.COLORMAP_PARULA, "Magma": cv2.COLORMAP_MAGMA, "Inferno": cv2.COLORMAP_INFERNO,
                      "Plasma": cv2.COLORMAP_PLASMA, "Viridis": cv2.COLORMAP_VIRIDIS, "Cividis": cv2.COLORMAP_CIVIDIS,
                      "Twilight": cv2.COLORMAP_TWILIGHT, "Twilight Shifted": cv2.COLORMAP_TWILIGHT_SHIFTED,
                      "Turbo": cv2.COLORMAP_TURBO, "Deep Green": cv2.COLORMAP_DEEPGREEN}
        self.colormap = self.cmaps['Viridis']

        # Action Widgets
        self.action_lbl = func.BackgroundLabel(text="Actions", size_hint=(1, 0.1),
                                               halign='center', valign='center', font_size=self.home.font)
        self.drag_btn = func.RoundedButton(text="Drag Mode", size_hint=(1, 0.1),
                                           halign='center', valign='center', font_size=self.home.font)
        self.drag_btn.bind(on_press=self.drag_mode)
        self.edit_btn = func.RoundedButton(text="Edit Mode", size_hint=(1, 0.1),
                                           halign='center', valign='center', font_size=self.home.font)
        self.edit_btn.bind(on_press=self.edit_mode)
        self.action_widgets = [self.action_lbl, self.drag_btn, self.edit_btn]

        # Editing Mode widgets
        self.back_btn = func.RoundedButton(text="Back", size_hint=(1, 0.1),
                                           halign='center', valign='center', font_size=self.home.font)
        self.back_btn.bind(on_press=self.edit_mode)
        self.delete_line_btn = func.RoundedButton(text="Delete Last Line", size_hint=(1, 0.1),
                                                  font_size=self.home.font)
        self.delete_line_btn.bind(on_press=lambda x: self.tool.del_line())
        self.delete_point_btn = func.RoundedButton(text="Delete Last Point", size_hint=(1, 0.1),
                                                   font_size=self.home.font)
        self.delete_point_btn.bind(on_press=lambda x: self.tool.del_point())

        self.edit_widgets = [self.back_btn, self.delete_line_btn, self.delete_point_btn]
        self.current = []

        self.load_image()

    # Turn off ScatterLayout functionality that conflicts with ImageView functionality
    ScatterLayout.do_rotation = False
    ScatterLayout.do_scale = False

    def font_adapt(self, font):
        """
        Update editing mode button font sizes.

        Args:
            font (float): New font size
        """
        self.action_lbl.font_size = font
        self.drag_btn.font_size = font
        self.edit_btn.font_size = font
        self.back_btn.font_size = font
        self.delete_line_btn.font_size = font
        self.delete_point_btn.font_size = font
        if self.t_mode:
            self.tool.font_adapt(font)

    def load_image(self):
        """
        Creates UI Image element, loads it in viewer, and scales it to window size.
        """

        if self.f_type == "netcdf":
            self.netcdf_to_image()
            self.im = CoreImage(self.byte, ext='png')
            self.size = self.im.size
        elif self.f_type == "image":
            self.im = CoreImage(self.config[self.f_type])
            self.size = im.open(self.config[self.f_type]).size

        self.img = ui.image.Image(source="", texture=self.im.texture, size=self.size, pos=self.home.ids.view.pos)
        self.add_widget(self.img)

        # Begin at max size where you can see entire image
        w_size = self.home.ids.view.size
        i_size = self.bbox[1]
        if i_size[0] >= w_size[0] or i_size[1] >= w_size[1]:
            while self.bbox[1][0] >= w_size[0] or self.bbox[1][1] >= w_size[1]:
                mat = Matrix().scale(.9, .9, .9)
                self.apply_transform(mat)
        if i_size[0] <= w_size[0] and i_size[1] <= w_size[1]:
            while self.bbox[1][0] <= w_size[0] and self.bbox[1][1] <= w_size[1]:
                mat = Matrix().scale(1.1, 1.1, 1.1)
                self.apply_transform(mat)

        xco = w_size[0] / 2 - self.bbox[1][0] / 2
        self.pos = (xco, self.pos[1])

    def manage_tool(self, t_type):
        """
        Manages the creation and deletion of each tool.

        If a tool is currently loaded, remove it, switch cursor to an arrow
        and clean up. If not, load indicated tool and switches cursor to a cross.

        Args:
            t_type (str): Tool type: 'transect' or 'transect_marker'
        """

        if not self.t_mode:
            # Opens a new tool
            for w in self.action_widgets:
                self.sidebar.add_widget(w, 1)
            kivy.core.window.Window.set_system_cursor("crosshair")
            if t_type == "transect":
                self.tool = MultiTransect(home=self.home)
            elif t_type == "transect_marker":
                self.tool = MultiMarker(home=self.home)
            self.add_widget(self.tool)
            self.t_mode = True
        else:
            kivy.core.window.Window.set_system_cursor("arrow")
            self.remove_widget(self.tool)
            self.drag_btn.text = "Drag Mode"
            self.editing = False
            self.reset_sidebar()
            self.t_mode = False

    def reset_sidebar(self):
        """
        Revert sidebar to pre-tool state
        """
        while self.sidebar.children != self.og_sidebar:
            self.sidebar.remove_widget(self.sidebar.children[1])

    def edit_mode(self, *args):
        """
        Turns editing mode on if off and off if on.

        When editing mode is being turned on, all sidebar buttons below 'Actions' label are added to
        a temporary holding list and then replaced with editing buttons. When editing mode is turned
        off, the editing buttons are removed and the original buttons are returned to the sidebar.

        Args:
            *args: Unused arguments passed to method when called.
        """
        if self.editing:
            for i in self.edit_widgets:
                if i in self.sidebar.children:
                    self.sidebar.remove_widget(i)
            for i in reversed(self.current):
                if i not in self.sidebar.children:
                    self.sidebar.add_widget(i, 1)
            self.editing = False
        else:
            if self.dragging:  # Can't be in both dragging and editing mode
                self.drag_mode()
            self.current = self.sidebar.children[1:self.sidebar.children.index(self.action_lbl)]
            for i in self.current:
                if i in self.sidebar.children:
                    self.sidebar.remove_widget(i)
            for i in self.edit_widgets:
                if i not in self.sidebar.children:
                    self.sidebar.add_widget(i, 1)
            self.editing = True
        self.tool.change_dragging(self.editing)

    def drag_mode(self, *args):
        """
        Calls for transecting tool to turn dragging mode on if off and off if on

        Args:
            *args: Unused arguments passed to method when called.
        """
        if not self.dragging:
            self.drag_btn.text = "Transect Mode"
            kivy.core.window.Window.set_system_cursor("arrow")
            self.tool.change_dragging(True)
            self.dragging = True
        else:
            self.drag_btn.text = "Drag Mode"
            kivy.core.window.Window.set_system_cursor("crosshair")
            self.tool.change_dragging(False)
            self.dragging = False

    def update_settings(self, setting, value):
        """
        Updates app settings depending on context of setting

        Args:
            setting (str): Name of setting being changed
            value: New setting value of appropriate data type
        """
        if setting == "l_color":
            self.l_col = value
            if self.t_mode:
                self.tool.update_l_col(value)
        elif setting == "cir_size":
            self.cir_size = float(value)
            if self.t_mode:
                self.tool.update_c_size(float(value))
        elif setting == "contrast":
            if self.f_type == "netcdf":
                self.contrast = float(value)
                self.update_netcdf()
        elif setting == "colormap":
            if self.f_type == "netcdf":
                self.colormap = self.cmaps[value]
                self.update_netcdf()
        elif setting == "variable":
            if self.f_type == "netcdf":
                self.config['netcdf']['var'] = value
                self.update_netcdf()
        elif setting == "depth":
            if self.f_type == "netcdf" and self.config['netcdf']['z'] != 'N/A':
                self.config['netcdf']['z_val'] = value
                self.update_netcdf()

    def update_netcdf(self):
        """
        Reload netcdf image when netcdf data is changed.
        """
        self.netcdf_to_image()
        self.im = CoreImage(self.byte, ext='png')
        self.size = self.im.size
        self.img.texture = self.im.texture
        self.img.reload()

    def netcdf_to_image(self):
        """
        Creates image from NetCDF dataset defined in :attr:`nccut.filedisplay.FileDisplay.f_config`

        Normalizes data and then rescales it to between 0 and 255. Applies colormap and contrast settings
        and then calls for the creation of colorbar. Loads image into memory as io.BytesIO object so kivy
        can make image out of an array.
        """
        self.nc_data = func.sel_data(self.config['netcdf'])
        n_data = (self.nc_data - np.nanmin(self.nc_data)) / (np.nanmax(self.nc_data) - np.nanmin(self.nc_data))
        n_data = np.nan_to_num(n_data, nan=1)
        n_data = (n_data * 255).astype(np.uint8)
        img = cv2.applyColorMap(n_data, self.colormap)
        self.update_color_bar(self.get_color_bar())

        # Applies contrast settings
        img = self.apply_contrast(img, self.contrast)
        is_success, img_b = cv2.imencode(".png", img)
        self.byte = io.BytesIO(img_b)

    def update_color_bar(self, colorbar):
        """
        Adds colorbar image to settings bar.

        Args:
            colorbar: kivy.uix.image.Image, colorbar graphic
        """
        if len(self.home.ids.colorbar.children) != 0:
            self.home.ids.colorbar.remove_widget(self.home.ids.colorbar.children[0])
        self.home.ids.colorbar.add_widget(colorbar)

    def get_color_bar(self):
        """
        Create color bar image according to colormap and dataset
        """
        c_arr = (np.arange(0, 256) * np.ones((10, 256))).astype(np.uint8)
        c_bar = cv2.applyColorMap(c_arr, self.colormap)
        c_bar = cv2.cvtColor(c_bar, cv2.COLOR_BGR2RGB)
        plt.figure(figsize=(10, 1))
        plt.imshow(c_bar)

        ax = plt.gca()
        ax.get_yaxis().set_visible(False)
        lab_arr = np.linspace(np.nanmin(self.nc_data), np.nanmax(self.nc_data), 256)
        lab_arr = ["{:.2e}".format(elem) for elem in lab_arr]
        ticks = [0, 50, 100, 150, 200, 250]
        ax.set_xticks(ticks=ticks, labels=[lab_arr[i] for i in ticks], fontsize=20)
        ax.xaxis.label.set_color('white')
        ax.tick_params(axis='x', colors='white')
        temp = io.BytesIO()
        plt.savefig(temp, facecolor=(0.2, 0.2, 0.2), bbox_inches='tight', format="png")
        temp.seek(0)
        plt.close()
        plot = ui.image.Image(source="", texture=CoreImage(io.BytesIO(temp.read()), ext="png").texture)
        return plot

    def apply_contrast(self, data, contrast):
        """
        Perform contrast level adjustment to data

        Args:
            data: 2D array to which to apply contrast change
            contrast (int): Desired contrast value

        Return:
            Data with contrast adjusted
        """
        f = 131 * (contrast + 127) / (127 * (131 - contrast))
        alpha_c = f
        gamma_c = 127 * (1 - f)
        out = cv2.addWeighted(data, alpha_c, data, 0, gamma_c)
        return out

    def flip_vertically(self):
        """
        Flips display vertically
        """
        m = Matrix()
        m.set(array=[
            [1.0, 0.0, 0.0, 0.0],
            [0.0, -1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0]]
        )
        self.apply_transform(m, anchor=self.center)

    def flip_horizontally(self):
        """
        Flips display horizontally
        """
        m = Matrix()
        m.set(array=[
            [-1.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0]]
        )
        self.apply_transform(m, anchor=self.center)

    def rotate(self):
        """
        Rotates display 45 degrees counterclockwise
        """
        self.rotation += 45

    def on_touch_down(self, touch):
        """
        If touch is of a scrolling type zoom in or out of the image.

        Args:
            touch: MouseMotionEvent, see kivy docs for details
        """
        # Scroll to zoom
        if self.collide_point(*touch.pos):
            if touch.is_mouse_scrolling:
                if touch.button == 'scrolldown':
                    if self.scale < 10:
                        mat = Matrix().scale(1.1, 1.1, 1.1)
                        self.apply_transform(mat, anchor=touch.pos)
                elif touch.button == 'scrollup':
                    if self.scale > 0.1:
                        mat = Matrix().scale(.9, .9, .9)
                        self.apply_transform(mat, anchor=touch.pos)
            else:
                super(FileDisplay, self).on_touch_down(touch)



"""
Graphics for a single transect whether drawn by user or used to hold data by marker tool.

Draws circle at each user click point. When one point has been clicked, draws line from that
point to user cursor.
"""

import kivy.uix as ui
from kivy.graphics import Color, Ellipse, Line
from kivy.metrics import dp
from kivy.uix.label import Label
from kivy.core.window import Window


class SingleTransect(ui.widget.Widget):
    """
    Graphics for a single transect whether drawn by user or used to hold data by marker tool.

    Draws circle at each user click point. When one point has been clicked, draws line from that
    point to user cursor.

    Attributes:
        line: Line object which holds endpoint data
        c1: Ellipse object for first point clicked
        c2: Ellipse object for second point clicked
        number: Label, reference to number label
        circles (int): Number of circles drawn, ie points clicked by user
        home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        size: 2 element array of ints, Size of widget
        pos: 2 element array of ints, Position of widget
        l_color: kivy.graphics.Color, Color to use for graphics
        c_size (tuple): 2 element tuple of floats that defines size of circles and other graphics
        line_width (float): Float, width of lines
    """
    def __init__(self, home, **kwargs):
        """
        Sets initial settings for transect according to attributes of HomeScreen root

        Args:
            home: Reference to root :class:`nccut.homescreen.HomeScreen` instance
        """
        super(SingleTransect, self).__init__(**kwargs)
        self.line = None
        self.c1 = None
        self.c2 = None
        self.number = None
        self.circles = 0
        self.home = home
        self.size = self.home.display.size
        self.pos = self.home.display.pos
        color = self.home.display.l_col
        if color == "Blue":
            self.l_color = Color(0.28, 0.62, 0.86)
        elif color == "Green":
            self.l_color = Color(0.39, 0.78, 0.47)
        elif color == "Orange":
            self.l_color = Color(0.74, 0.42, 0.13)
        size = self.home.display.cir_size
        self.c_size = (dp(size), dp(size))
        self.line_width = dp(size / 5)

    def update_l_col(self, color):
        """
        Updates the line color for the transect and redraws canvas elements

        Args:
            color (str): New color value: 'Blue', 'Green' or 'Orange'
        """
        if color == "Blue":
            self.l_color = Color(0.28, 0.62, 0.86)
        elif color == "Green":
            self.l_color = Color(0.39, 0.78, 0.47)
        elif color == "Orange":
            self.l_color = Color(0.74, 0.42, 0.13)
        graphics = [self.c1, self.c2, self.line]
        for g in graphics:
            if g:
                self.canvas.remove(g)
        self.canvas.add(self.l_color)
        for g in graphics:
            if g:
                self.canvas.add(g)

    def update_c_size(self, value):
        """
        Updates graphic sizes according to the new value.

        Args:
            value (float): New graphics size
        """
        self.line_width = dp(value / 5)
        self.line.width = self.line_width
        for c in [self.c1, self.c2]:
            if c:
                c.size = (dp(value), dp(value))
                c.pos = (c.pos[0] + self.c_size[0] / 2 - dp(value) / 2, c.pos[1] + self.c_size[1] / 2 - dp(value) / 2)
        self.number.font_size = dp(value) * 2
        self.number.pos = (self.number.pos[0] + self.c_size[0] - dp(value),
                           self.number.pos[1] + self.c_size[1] - dp(value))
        self.c_size = (dp(value), dp(value))

    def on_touch_down(self, touch):
        """
        Draws points and lines depending on which parts of the transect have already been drawn.

        Args:
            touch: MouseMotionEvent, see kivy docs for details
        """
        if self.home.ids.view.collide_point(*self.home.ids.view.to_widget(*self.to_window(*touch.pos))):
            if self.circles == 0:  # If first click, add number, one circle, and draw line between point and cursor
                self.circles += 1
                self.number = Label(text=str(len(self.parent.children)),
                                    pos=(touch.x + self.c_size[0], touch.y + self.c_size[0]),
                                    font_size=self.c_size[0] * 2)
                self.add_widget(self.number)
                with self.canvas:
                    # Add points and start line
                    Color(self.l_color.r, self.l_color.g, self.l_color.b)
                    self.line = Line(points=[], width=self.line_width, group='end')
                    self.c1 = Ellipse(pos=(touch.x - self.c_size[0] / 2, touch.y - self.c_size[1] / 2),
                                      size=self.c_size, group='start')
                    self.line.points = (touch.x, touch.y)
                # Bind line drawing to any time user mouse position changes
                Window.bind(mouse_pos=self.draw_line)

            elif self.circles == 1:  # If second click draw another circle and draws line between circles
                self.circles += 1
                if len(self.line.points) != 2:  # Stops drawing line between first circle and cursor
                    self.line.points.pop()
                    self.line.points.pop()
                if [touch.x, touch.y] == self.line.points:  # If clicked the same point as before, do nothing
                    return
                with self.canvas:  # Second circle and sets line between the two circles.
                    self.c2 = Ellipse(pos=(touch.x - self.c_size[0] / 2, touch.y - self.c_size[1] / 2),
                                      size=self.c_size, group='end')
                    self.line.points += (touch.x, touch.y)

    def draw_line(self, instance, pos):
        """
        When one point is clicked and user mouse is just hovering, continuously update line end point
        to be at mouse cursor

        Updates anytime cursor moves. Does not draw if tool in dragging mode or cursor not in viewing window.

        Args:
            instance: WindowSDL instance, current window loaded (not used by method)
            pos (tuple): 2 element tuple of floats, x and y coord of cursor position
        """
        if not self.parent.dragging:
            if self.home.ids.view.collide_point(*self.home.ids.view.to_widget(*pos)):
                mouse = self.to_widget(*pos)
                if self.size[0] >= mouse[0] >= 0 and self.size[1] >= mouse[1] >= 0:
                    # If cursor within image bounds
                    if self.circles == 1:
                        if len(self.line.points) != 2:
                            # If already drawn previously, remove end point
                            self.line.points.pop()
                            self.line.points.pop()
                        with self.canvas:

                            self.line.points += self.to_widget(pos[0], pos[1])
        else:
            # If in dragging mode, don't draw and erase line
            if self.parent.lines[-1] == self and self.circles == 1:
                with self.canvas:
                    self.line.points = self.line.points[0:2]

    def del_point(self):
        """
        Delete most recently clicked point and reset
        """
        self.canvas.remove_group('end')
        self.circles -= 1
        self.line.points = self.line.points[:2]
        self.canvas.add(self.line)
        Window.bind(mouse_pos=self.draw_line)
